"""
Driver for running model on design of experiments cases using OpenTURNS sampling methods
"""
import numpy as np

from openmdao.api import DOEDriver
from openmdao.drivers.doe_generators import DOEGenerator

OPENTURNS_NOT_INSTALLED = False
try:
    import openturns as ot
except ImportError:
    OPENTURNS_NOT_INSTALLED = True


class OpenturnsMonteCarloDOEGenerator(DOEGenerator):
    LIMIT = 1e12

    def __init__(self, n_samples=10, dist=None):
        super(OpenturnsMonteCarloDOEGenerator, self).__init__()

        self.n_samples = n_samples
        self.distribution = dist
        self.called = False

    def __call__(self, uncertain_vars, model=None):
        if self.distribution is None:
            dists = []
            for name, meta in uncertain_vars.items():
                size = meta["size"]
                meta_low = meta["lower"]
                meta_high = meta["upper"]
                for j in range(size):
                    if isinstance(meta_low, np.ndarray):
                        p_low = meta_low[j]
                    else:
                        p_low = meta_low
                    p_low = max(p_low, -self.LIMIT)

                    if isinstance(meta_high, np.ndarray):
                        p_high = meta_high[j]
                    else:
                        p_high = meta_high
                    p_high = min(p_high, self.LIMIT)

                    dists.append(ot.Uniform(p_low, p_high))
            self.distribution = ot.ComposedDistribution(dists)
        else:
            size = 0
            for name, meta in uncertain_vars.items():
                size += meta["size"]
            if (size) != (self.distribution.getDimension()):
                raise RuntimeError(
                    "Bad distribution dimension: should be equal to uncertain variables size {} "
                    ", got {}".format(size, self.distribution.getDimension())
                )
        samples = self.distribution.getSample(self.n_samples)
        self._cases = np.array(samples)
        self.called = True
        sample = []
        for i in range(self._cases.shape[0]):
            j = 0
            for name, meta in uncertain_vars.items():
                size = meta["size"]
                sample.append((name, self._cases[i, j : j + size]))
                j += size
            yield sample

    def get_cases(self):
        if not self.called:
            raise RuntimeError("Have to run the driver before getting cases")
        return self._cases


class OpenturnsDOEDriver(DOEDriver):
    """
    Baseclass for OpenTURNS design-of-experiments Drivers
    """

    def __init__(self, **kwargs):
        super(OpenturnsDOEDriver, self).__init__()

        self.options.declare(
            "distribution",
            types=ot.ComposedDistribution,
            default=None,
            allow_none=True,
            desc="Joint distribution of uncertain variables",
        )
        self.options.declare(
            "n_samples", types=int, default=2, desc="number of sample to generate"
        )
        self.options.update(kwargs)

        self.options["generator"] = OpenturnsMonteCarloDOEGenerator(
            self.options["n_samples"], self.options["distribution"]
        )

    def _set_name(self):
        self._name = "OpenTURNS_DOE_MonteCarlo"

    def get_cases(self):
        return self.options["generator"].get_cases()



import numpy as np
import traceback

from openmdao.core.driver import Driver, RecordingDebugging
from openmdao.core.analysis_error import AnalysisError

EGOBOX_NOT_INSTALLED = False
try:
    import egobox as egx
    from egobox import Egor
except ImportError:
    EGOBOX_NOT_INSTALLED = True


def to_list(l, size):
    if not (isinstance(l, np.ndarray) or isinstance(l, list)):
        return [l] * size
    diff_len = len(l) - size
    if diff_len > 0:
        return l[0:size]
    elif diff_len < 0:
        return [l[0]] * size
    else:
        return l


class EgoboxEgorDriver(Driver):
    """OpenMDAO driver for egobox optimizer"""

    def __init__(self, **kwargs):
        """Initialize the driver with the given options."""
        super(EgoboxEgorDriver, self).__init__(**kwargs)

        if EGOBOX_NOT_INSTALLED:
            raise RuntimeError("egobox library is not installed.")

        # What we support
        self.supports["optimization"] = True
        self.supports["inequality_constraints"] = True
        self.supports["linear_constraints"] = True
        self.supports["integer_design_vars"] = True

        # What we don't support
        self.supports["equality_constraints"] = False
        self.supports["two_sided_constraints"] = False
        self.supports["multiple_objectives"] = False
        self.supports["active_set"] = False
        self.supports["simultaneous_derivatives"] = False
        self.supports["total_jac_sparsity"] = False
        self.supports["gradients"] = False
        self.supports._read_only = True

        self.opt_settings = {}

    def _declare_options(self):
        self.options.declare(
            "optimizer",
            default="EGOR",
            values=["EGOR"],
            desc="Name of optimizer to use",
        )

    def _setup_driver(self, problem):
        super(EgoboxEgorDriver, self)._setup_driver(problem)

        self.comm = None

    def run(self):
        model = self._problem().model

        self.iter_count = 0
        self.name = f"egobox_optimizer_{self.options['optimizer'].lower()}"

        # Initial Run
        with RecordingDebugging(self.name, self.iter_count, self) as rec:
            # Initial Run
            model.run_solve_nonlinear()
            rec.abs = 0.0
            rec.rel = 0.0
        self.iter_count += 1

        # Format design variables to suit segomoe implementation
        self.xspecs = self._initialize_vars(model)

        # Format constraints to suit segomoe implementation
        self.n_cstr = self._initialize_cons()

        # Format option dictionary to suit Egor implementation
        optim_settings = {
            "cstr_tol": 1e-6,
        }
        n_iter = self.opt_settings["maxiter"]
        optim_settings.update(
            {k: v for k, v in self.opt_settings.items() if k != "maxiter"}
        )

        dim = 0
        for name, meta in self._designvars.items():
            dim += meta["size"]
        if dim > 10:
            self.optim_settings["kpls_dim"] = 3

        # Instanciate a SEGO optimizer
        egor = Egor(
            self._objfunc,
            xspecs=self.xspecs,
            n_cstr=self.n_cstr,
            **optim_settings,
        )

        # Run the optim
        res = egor.minimize(n_iter=n_iter)

        # Set optimal parameters
        i = 0
        for name, meta in self._designvars.items():
            size = meta["size"]
            self.set_design_var(name, res.x_opt[i : i + size])
            i += size

        with RecordingDebugging(self.name, self.iter_count, self) as rec:
            model.run_solve_nonlinear()
            rec.abs = 0.0
            rec.rel = 0.0
        self.iter_count += 1

        return True

    def _initialize_vars(self, model):
        dvs_int = {}
        for name, meta in self._designvars.items():
            infos = model.get_io_metadata(includes=name)
            for absname in infos:
                if name == infos[absname]["prom_name"] and (
                    infos[absname]["tags"] & {"wop:int"}
                ):
                    dvs_int[name] = egx.XType(egx.XType.INT)

        variables = []
        desvars = self._designvars
        for name, meta in desvars.items():
            vartype = dvs_int.get(name, egx.XType(egx.XType.FLOAT))
            if meta["size"] > 1:
                if np.isscalar(meta["lower"]):
                    variables += [
                        egx.XSpec(vartype, [meta["lower"], meta["upper"]])
                        for i in range(meta["size"])
                    ]
                else:
                    variables += [
                        egx.XSpec(vartype, [meta["lower"], meta["upper"]])
                        for i in range(meta["size"])
                    ]
            else:
                variables += [egx.XSpec(vartype, [meta["lower"], meta["upper"]])]
        return variables

    def _initialize_cons(self, eq_tol=None, ieq_tol=None):
        """Format OpenMDAO constraints to suit EGOR implementation

        Parameters
        ----------
        eq_tol: dict
            Dictionary to define specific tolerance for eq constraints
            {'[groupName]': [tol]} Default tol = 1e-5
        """
        con_meta = self._cons

        self.ieq_cons = {
            name: con for name, con in con_meta.items() if not con["equals"]
        }

        # Inequality constraints
        n_cstr = 0
        for name in self.ieq_cons.keys():
            meta = con_meta[name]
            size = meta["size"]
            # Bounds - double sided is supported
            lower = to_list(meta["lower"], size)
            upper = to_list(meta["upper"], size)
            for k in range(size):
                if (lower[k] is None or lower[k] < -1e29) and upper[k] == 0.0:
                    n_cstr += 1
                else:
                    raise ValueError(
                        f"Constraint {lower[k]} < g(x) < {upper[k]} not handled by Egor driver"
                    )
        return n_cstr

    def _objfunc(self, points):
        """
        Function that evaluates and returns the objective function and the
        constraints. This function is called by SEGOMOE

        Parameters
        ----------
        point : numpy.ndarray
            point to evaluate

        Returns
        -------
        func_dict : dict
            Dictionary of all functional variables evaluated at design point.
        fail : int
            0 for successful function evaluation
            1 for unsuccessful function evaluation
        """
        res = np.zeros((points.shape[0], 1 + self.n_cstr))
        model = self._problem().model

        for k, point in enumerate(points):
            try:
                # Pass in new parameters
                i = 0

                for name, meta in self._designvars.items():
                    size = meta["size"]
                    self.set_design_var(name, point[i : i + size])
                    i += size

                # Execute the model
                with RecordingDebugging(
                    self.options["optimizer"], self.iter_count, self
                ) as _:
                    self.iter_count += 1
                    try:
                        model.run_solve_nonlinear()

                    # Let the optimizer try to handle the error
                    except AnalysisError:
                        model._clear_iprint()

                # Get the objective function evaluation - single obj support
                for obj in self.get_objective_values().values():
                    res[k, 0] = obj

                # Get the constraint evaluations
                j = 1
                for con_res in self.get_constraint_values().values():
                    # Make sure con_res is array_like
                    con_res = to_list(con_res, 1)
                    # Perform mapping
                    for i, _ in enumerate(con_res):
                        res[k, j + i] = con_res[i]
                    j += 1

            except Exception as msg:
                tb = traceback.format_exc()
                print("Exception: %s" % str(msg))
                print(70 * "=", tb, 70 * "=")

        return res



"""
Author: Remi Lafage <remi.lafage@onera.fr>

This package is distributed under Apache 2 license.
"""

from setuptools import setup

from os import path
from io import open
from openmdao_extensions import __version__

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

CLASSIFIERS = """
Development Status :: 5 - Production/Stable
Intended Audience :: Science/Research
Intended Audience :: Developers
License :: OSI Approved :: Apache Software License
Programming Language :: Python
Programming Language :: Python :: 3
Topic :: Software Development
Topic :: Scientific/Engineering
Operating System :: Microsoft :: Windows
Operating System :: Unix
Operating System :: MacOS
"""

metadata = dict(
    name="openmdao_extensions",
    version=__version__,
    description="Additional solvers and drivers for OpenMDAO framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Rémi Lafage",
    author_email="remi.lafage@onera.fr",
    license="Apache License, Version 2.0",
    classifiers=[_f for _f in CLASSIFIERS.split("\n") if _f],
    packages=["openmdao_extensions"],
    install_requires=["openmdao>=3.0.0"],
    python_requires=">=3.7",
    zip_safe=True,
    url="https://github.com/OneraHub/openmdao_extensions",
    download_url="https://github.com/OneraHub/openmdao_extensions/releases",
    keywords="openmdao openmdao_driver openmdao_nl_solver",
    entry_points={
        "openmdao_driver": [
            "egoboxegordriver = openmdao_extensions.egobox_egor_driver:EgoboxEgorDriver",
            "onerasegodriver = openmdao_extensions.onera_sego_driver:OneraSegoDriver",
            "openturnsdoedriver = openmdao_extensions.openturns_doe_driver:OpenturnsDOEDriver",
            "salibdoedriver = openmdao_extensions.salib_doe_driver:SalibDOEDriver",
            "smtdoedriver = openmdao_extensions.smt_doe_driver:SmtDOEDriver",
        ],
        "openmdao_nl_solver": [
            "recklessnonlinearblockgs = openmdao_extensions.reckless_nonlinear_block_gs:RecklessNonlinearBlockGS"
        ],
    },
)

setup(**metadata)
